/**
  *  \file planetscentral/format/simpacker.cc
  *  \brief Format - Simulator Packer
  *
  *  This re-implements sim-io.cc somehow, because that has some ugly
  *  interdependencies we do not want.
  */

#include <stdexcept>
#include <cstddef>
#include <cstring>
#include "planetscentral/format/simpacker.h"
#include "game/struct.h"
#include "int/hash.h"
#include "int/array.h"
#include "planetscentral/dbclient/connection.h"

static const int MAX_VERSION  = 5;
static const int MAX_OBJ_SIZE = 67;

static const uint8_t record_sizes[] = {
    51, 53, 53, 57, 65, 67
};

static const std::size_t MAGIC_V0_LEN = 6;
static const std::size_t MAGIC_V1_LEN = 8;

static const char magic_v0[] = { 'C','C','s','i','m',26 };
static const char magic_v1[] = { 'C','C','b','s','i','m','0',26 };

CTASSERT(MAX_OBJ_SIZE == TSimShipData::size);
CTASSERT(MAX_OBJ_SIZE == TSimPlanetData::size);

static int
checkVersion(IntValue* obj, int existing)
{
    int my_version;
    int32_t flags = toIntResult(getHashElem(obj, "FLAGS"));
    if ((flags >> 16) != 0) {
        my_version = 5;
    } else if ((flags & 16 /*fl_RatingOverride*/) != 0) {
        my_version = 4;
    } else if (toIntResult(getHashElem(obj, "MISSION.INTERCEPT")) != 0) {
        my_version = 4;
    } else {
        my_version = 3;
    }

    return my_version > existing ? my_version : existing;
}

/******************************* SimPacker *******************************/

string_t
SimPacker::pack(IntValue* data)
{
    // Figure out key parameters
    IntValue* shipArray = getHashElem(data, "ships");
    IntValue* planet    = getHashElem(data, "planet");
    int num_ships = getArraySize(shipArray);

    // Figure out version to write
    int version = checkVersion(planet, 3);
    for (int i = 0; i < num_ships; ++i) {
        version = checkVersion(getArrayElem(shipArray, i), version);
    }

    // Store header
    string_t result;
    char header[MAGIC_V1_LEN + 2];
    std::memcpy(header, magic_v1, MAGIC_V1_LEN);
    header[6] = '0' + version - 1;

    if (planet != 0) {
        storeUint16(header + MAGIC_V1_LEN, num_ships | 0x8000);
    } else {
        storeUint16(header + MAGIC_V1_LEN, num_ships);
    }
    result.append(header, sizeof(header));

    // Store ships
    for (int i = 0; i < num_ships; ++i) {
        char buffer[MAX_OBJ_SIZE];
        TSimShipData sh;
        packShip(sh, getArrayElem(shipArray, i), version);
        storeStructure(buffer, sh);
        result.append(buffer, record_sizes[version]);
    }

    // Store planet
    if (planet != 0) {
        char buffer[MAX_OBJ_SIZE];
        TSimPlanetData pl;
        packPlanet(pl, planet, version);
        storeStructure(buffer, pl);
        result.append(buffer, record_sizes[version]);
    }
    return result;
}

IntValue*
SimPacker::unpack(const string_t& data)
{
    // Figure out version and place initial cursor
    string_t::size_type pos;
    int version;
    if (data.size() >= MAGIC_V0_LEN
        && data.compare(0, MAGIC_V0_LEN, magic_v0, MAGIC_V0_LEN) == 0)
    {
        version = 0;
        pos     = MAGIC_V0_LEN;
    } else if (data.size() >= MAGIC_V1_LEN
               && data.compare(0, MAGIC_V0_LEN, magic_v1, MAGIC_V0_LEN) == 0 /* MAGIC_V0_LEN is correct here! */
               && data[MAGIC_V0_LEN] >= '0'
               && data[MAGIC_V0_LEN] < '0' + MAX_VERSION)
    {
        version = data[MAGIC_V0_LEN] - '0' + 1;
        pos     = MAGIC_V1_LEN;
    } else {
        throw std::runtime_error("File format error");
    }

    // Check object count
    TInt16 count;
    if (!getStructure(data, pos, count)) {
        throw std::runtime_error("File format error");
    }
    uint16_t num_ships = (count.value & 0x7FFF);
    bool has_planet = (count.value & 0x8000) != 0;

    // Make room for output
    Ptr<IntHashData> hd = new IntHashData();

    // Read Ships
    Ptr<IntArrayData> sa = new IntArrayData();
    for (uint16_t i = 0; i < num_ships; ++i) {
        // Fetch data
        char obj_buf[MAX_OBJ_SIZE];
        if (pos >= data.size() || data.size() - pos < record_sizes[version]) {
            throw std::runtime_error("File format error");
        }
        std::memcpy(obj_buf, &data[pos], record_sizes[version]);
        pos += record_sizes[version];

        // Parse it
        TSimShipData data;
        data.aggressiveness = -1;
        data.mass           = 100;
        data.flags          = 0;
        data.flak_r         = 0;
        data.flak_c         = 0;
        data.xa_id          = 0;
        getPartialStructure(obj_buf, &data, data.parser, 0, record_sizes[version]);

        // Generate output
        sa->content.pushNew(describeShip(data, version));
    }
    sa->addDimension(sa->content.size());
    hd->setNew("ships", new IntArray(sa));

    if (has_planet) {
        // Fetch data
        char obj_buf[MAX_OBJ_SIZE];
        if (pos >= data.size() || data.size() - pos < record_sizes[version]) {
            throw std::runtime_error("File format error");
        }
        std::memcpy(obj_buf, &data[pos], record_sizes[version]);
        pos += record_sizes[version];

        // Parse it
        TSimPlanetData data;
        data.aggressiveness = -1;
        data._pad5          = 100;
        data.flags          = 0;
        data.flak_r         = 0;
        data.flak_c         = 0;
        data._pad6          = 0;
        getPartialStructure(obj_buf, &data, data.parser, 0, record_sizes[version]);

        hd->setNew("planet", describePlanet(data, version));
    }

    return new IntHash(hd);
}

IntValue*
SimPacker::describeShip(const TSimShipData& sh, int version)
{
    Ptr<IntHashData> hd = new IntHashData();

    /* Version dependencies */
    int32_t agg   = version <= 0 ? -1 /*agg_Kill*/ : sh.aggressiveness;
    int32_t flags = (version < 3
                     ? 0
                     : version <= 3
                     ? uint16_t(sh.flags) & ~16 /*fl_RatingOverride*/
                     : version < 5
                     ? uint16_t(sh.flags)
                     : uint16_t(sh.flags) + 65536*sh.flags2);

    /* Map fighters/torps to one, as we do for ships */
    int32_t aux_type = (sh.vcro._torp_launcher_count > 0 && sh.vcro.torp_type > 0 && sh.vcro.torp_type <= NUM_TORPS
                        ? sh.vcro.torp_type
                        : sh.vcro.bay_count > 0
                        ? NUM_TORPS+1
                        : 0);
    int32_t aux_count = (aux_type > 0 && aux_type <= NUM_TORPS
                         ? sh.vcro._torp_launcher_count
                         : aux_type == NUM_TORPS+1
                         ? sh.vcro.bay_count
                         : 0);

    /* Build result */
    hd->setNew("AGGRESSIVENESS",    makeIntValue(agg));
    hd->setNew("AUX",               makeIntValue(aux_type));
    hd->setNew("AUX.AMMO",          makeIntValue(sh.vcro._ammo));
    hd->setNew("AUX.COUNT",         makeIntValue(aux_count));
    hd->setNew("BEAM",              makeIntValue(sh.vcro.beam_type));
    hd->setNew("BEAM.COUNT",        makeIntValue(sh.vcro.beam_count));
    hd->setNew("CREW",              makeIntValue(sh.vcro.crew));
    hd->setNew("DAMAGE",            makeIntValue(sh.vcro.damage));
    hd->setNew("ENGINE",            makeIntValue(sh.engine_type));
    hd->setNew("FCODE",             makeStringValue(unpackString(sh.fcode, sizeof(sh.fcode))));
    hd->setNew("FLAGS",             makeIntValue(flags));
    hd->setNew("HULL",              makeIntValue(sh.hull_type));
    hd->setNew("ID",                makeIntValue(sh.vcro.id));
    hd->setNew("LEVEL",             makeIntValue(sh.vcro.experience_level));
    hd->setNew("MASS",              makeIntValue(sh.mass));
    hd->setNew("MISSION.INTERCEPT", makeIntValue(sh.xa_id));
    hd->setNew("NAME",              makeStringValue(unpackString(sh.vcro.name, sizeof(sh.vcro.name))));
    hd->setNew("OWNER",             makeIntValue(sh.vcro.owner));
    hd->setNew("RATING.C",          makeIntValue(sh.flak_c));
    hd->setNew("RATING.R",          makeIntValue(sh.flak_r));
    hd->setNew("SHIELD",            makeIntValue(sh.shield));

    return new IntHash(hd);
}

IntValue*
SimPacker::describePlanet(const TSimPlanetData& pl, int version)
{
    /* Derived data */
    int beam_tech    = pl.beam_tech >= 0 && pl.beam_tech <= NUM_BEAMS ? pl.beam_tech : 0;
    int torp_tech    = pl.torp_tech >= 0 && pl.torp_tech <= NUM_TORPS ? pl.torp_tech : 0;
    int base_defense = pl.base_defense >= 0 && pl.base_defense <= 1000 ? pl.base_defense : 0;
    int defense      = pl.defense >= 0 && pl.defense < 1000 ? pl.defense : 0;
    int32_t flags = (version < 3
                     ? 0
                     : version <= 3
                     ? uint16_t(pl.flags) & ~16 /*fl_RatingOverride*/
                     : version < 5
                     ? uint16_t(pl.flags)
                     : uint16_t(pl.flags) + 65536*pl.flags2);

    Ptr<IntArrayData> base_ammo = new IntArrayData();
    for (int i = 0; i < NUM_TORPS; ++i) {
        int this_torps = (version > 1
                          ? pl.base_torps[i]
                          : torp_tech == i+1
                          ? pl.base_torps_old
                          : 0);
        base_ammo->content.pushNew(makeIntValue(this_torps));
    }
    base_ammo->content.pushNew(makeIntValue(pl.base_fighters));
    base_ammo->addDimension(base_ammo->content.size());

    /* Build result */
    Ptr<IntHashData> hd = new IntHashData();
    hd->setNew("AGGRESSIVENESS",    makeIntValue(-1));           // Not editable! --- FIXME: not even in the data!
    hd->setNew("DAMAGE",            makeIntValue(0));            // Not editable!
    hd->setNew("DEFENSE",           makeIntValue(defense));
    hd->setNew("DEFENSE.BASE",      makeIntValue(base_defense));
    hd->setNew("FCODE",             makeStringValue(unpackString(pl.fcode, sizeof(pl.fcode))));
    hd->setNew("FLAGS",             makeIntValue(flags));
    hd->setNew("ID",                makeIntValue(pl.id));
    hd->setNew("LEVEL",             makeIntValue(pl.experience_level));
    hd->setNew("OWNER",             makeIntValue(pl.owner));
    hd->setNew("RATING.C",          makeIntValue(pl.flak_c));
    hd->setNew("RATING.R",          makeIntValue(pl.flak_r));
    hd->setNew("SHIELD",            makeIntValue(100));          // Not editable!
    hd->setNew("STORAGE.AMMO",      new IntArray(base_ammo));
    hd->setNew("TECH.BEAM",         makeIntValue(beam_tech));
    hd->setNew("TECH.TORPEDO",      makeIntValue(torp_tech));

    return new IntHash(hd);
}

void
SimPacker::packShip(TSimShipData& sh, IntValue* p, int version)
{
    (void) version;

    // Derived values
    int aux_type  = toIntResult(getHashElem(p, "AUX"));
    int aux_count = toIntResult(getHashElem(p, "AUX.COUNT"));
    bool fighters = aux_type == NUM_TORPS+1;
    int32_t flags = toIntResult(getHashElem(p, "FLAGS"));

    // Pack it
    packString(sh.vcro.name, sizeof(sh.vcro.name), toStringResult(getHashElem(p, "NAME")));
    sh.vcro.damage               = toIntResult(getHashElem(p, "DAMAGE"));
    sh.vcro.crew                 = toIntResult(getHashElem(p, "CREW"));
    sh.vcro.id                   = toIntResult(getHashElem(p, "ID"));
    sh.vcro.owner                = toIntResult(getHashElem(p, "OWNER"));
    sh.vcro.race_or_zero         = 0;
    sh.vcro.picture              = 0;
    sh.vcro.hull_or_zero         = 0;
    sh.vcro.beam_type            = toIntResult(getHashElem(p, "BEAM"));
    sh.vcro.beam_count           = toIntResult(getHashElem(p, "BEAM.COUNT"));
    sh.vcro.experience_level     = toIntResult(getHashElem(p, "LEVEL"));
    sh.vcro.bay_count            = fighters ? aux_count : 0;
    sh.vcro.torp_type            = !fighters ? aux_type : 0;
    sh.vcro._ammo                = toIntResult(getHashElem(p, "AUX.AMMO"));
    sh.vcro._torp_launcher_count = !fighters ? aux_count : 0;
    sh.engine_type               = toIntResult(getHashElem(p, "ENGINE"));
    sh.hull_type                 = toIntResult(getHashElem(p, "HULL"));
    sh.shield                    = toIntResult(getHashElem(p, "SHIELD"));
    packString(sh.fcode, sizeof(sh.fcode), toStringResult(getHashElem(p, "FCODE")));
    sh.aggressiveness            = toIntResult(getHashElem(p, "AGGRESSIVENESS"));
    sh.mass                      = toIntResult(getHashElem(p, "MASS"));
    sh.flags                     = flags;
    sh.flak_r                    = toIntResult(getHashElem(p, "RATING.R"));
    sh.flak_c                    = toIntResult(getHashElem(p, "RATING.C"));
    sh.xa_id                     = toIntResult(getHashElem(p, "MISSION.INTERCEPT"));
    sh.flags2                    = flags >> 16;
}

void
SimPacker::packPlanet(TSimPlanetData& pl, IntValue* p, int version)
{
    // The minimum version we ever write is 3, so no need
    // to pack for versions 1 or lower that have different storage for torps.
    (void) version;

    // Prepare
    int32_t flags = toIntResult(getHashElem(p, "FLAGS"));
    IntValue* base_ammo = getHashElem(p, "STORAGE.AMMO");

    // Pack it
    for (int i = 0; i < NUM_TORPS; ++i) {
        pl.base_torps[i] = toIntResult(getArrayElem(base_ammo, i));
    }
    pl._pad0            = 0;
    pl.id               = toIntResult(getHashElem(p, "ID"));
    pl.owner            = toIntResult(getHashElem(p, "OWNER"));
    pl._pad1            = 0;
    pl.beam_tech        = toIntResult(getHashElem(p, "TECH.BEAM"));
    pl._pad2            = 0;
    pl.experience_level = toIntResult(getHashElem(p, "LEVEL"));
    pl.base_fighters    = toIntResult(getArrayElem(base_ammo, NUM_TORPS));
    pl._pad3            = 0;
    pl.base_torps_old   = 0;
    pl.torp_tech        = toIntResult(getHashElem(p, "TECH.TORPEDO"));
    pl.base_defense     = toIntResult(getHashElem(p, "DEFENSE.BASE"));
    pl.defense          = toIntResult(getHashElem(p, "DEFENSE"));
    pl.shield           = toIntResult(getHashElem(p, "SHIELD"));
    packString(pl.fcode, sizeof(pl.fcode), toStringResult(getHashElem(p, "FCODE")));
    pl.aggressiveness   = toIntResult(getHashElem(p, "AGGRESSIVENESS"));
    pl._pad5            = 0;
    pl.flags            = flags;
    pl.flak_r           = toIntResult(getHashElem(p, "RATING.R"));
    pl.flak_c           = toIntResult(getHashElem(p, "RATING.C"));
    pl._pad6            = 0;
    pl.flags2           = flags >> 16;
}
