/**
  *  \file planetscentral/mailout/transmitter.cc
  *  \brief Mailout: Mail Transmitter
  */

#include <stdexcept>
#include <stdio.h>
#include <stdarg.h>
#include "planetscentral/mailout/transmitter.h"
#include "planetscentral/mailout/mailout.h"
#include "planetscentral/mailout/db.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/osutil.h"
#include "cpluslib/format.h"
#include "util/except.h"
#include "planetscentral/net/md5.h"
#include "planetscentral/net/encoding.h"
#include "planetscentral/net/mimebuilder.h"
#include "planetscentral/mailout/template.h"
#include "io/textfile.h"
#include "io/streams.h"
#include "io/dirs.h"
#include "planetscentral/net/smtp.h"
#include "util/misc.h"

static string_t system_key;
static string_t web_root = "unconfigured";
static int32_t max_msg_age = 24*60*32;         // about one month

static string_t
getConfirmLink(string_t user, string_t userEmail)
{
    // Compute md5("$system_key,$user,$userEmail")
    unsigned char result[16];
    Md5Context ctx;
    ctx.init();
    ctx.update((const unsigned char*)system_key.data(), system_key.size());
    ctx.update((const unsigned char*)",", 1);
    ctx.update((const unsigned char*)user.data(), user.size());
    ctx.update((const unsigned char*)",", 1);
    ctx.update((const unsigned char*)userEmail.data(), userEmail.size());
    ctx.final(result);

    // Result is base64("$user,$result")
    return encodeBase64(user + "," + string_t((char*)result, 16));
}


static void
lockedPrintf(const char* fmt, ...)
{
    MutexGuard g(output_mutex);
    va_list va;
    va_start(va, fmt);
    vprintf(fmt, va);
    va_end(va);
}

/****************************** Transmitter ******************************/

/** Constructor. */
Transmitter::Transmitter()
    : work_sem(0)
{ }

/** Start transmitter. Initializes internal variables, and starts the
    transmitter thread. Call once. */
void
Transmitter::start()
{
    // Fetch current outgoing queue
    IntList_t out;
    {
        ReconnectingGuard g(database_rc);
        DbSubtree root(database_connection, MAIL_ROOT);
        root.intSetKey("sending").getAll(out);

        // Discard partially-prepared messages
        IntList_t partial;
        root.intSetKey("preparing").getAll(partial);
        for (uint32_t i = 0; i < partial.size(); ++i) {
            Message(root, partial[i], "preparing").remove();
        }
        lockedPrintf("[transmit] %d partial messages deleted\n", int(partial.size()));
    }

    // Send them all
    for (uint32_t i = 0; i < out.size(); ++i) {
        send(out[i]);
    }
    lockedPrintf("[transmit] %d items initially in queue\n", int(out.size()));

    // Start worker thread
    pthread_t handle;
    if (pthread_create(&handle, 0, entry, this) < 0) {
        perror("unable to create 'transmit' thread");
    }
}

/** Set template directory. Call during setup only. */
void
Transmitter::setTemplateDir(string_t dir)
{
    template_dir = dir;
}

/** Send a message. Call this after placing a message in MAIL_ROOT->sending. */
void
Transmitter::send(int32_t id)
{
    MutexGuard g(queue_mutex);
    work_queue.push_back(id);
    work_sem.up();
}

/** Notify status change of an address. */
void
Transmitter::notifyAddress(string_t /*adr*/)
{
    // Simple and stupid: just reconsider all messages
    runQueue();
}

/** Reconsider all messages for sending. Call periodically to clean up
    expired messages, or re-trigger sending postponed messages. */
void
Transmitter::runQueue()
{
    MutexGuard g(queue_mutex);
    while (!pend_queue.empty()) {
        work_queue.splice(work_queue.end(), pend_queue, pend_queue.begin());
        work_sem.up();
    }
}

/** Request confirmation for a user (CONFIRM command).
    Works by attempting to resolve the user Id into an address. This will
    trigger the request as a byproduct.
    \retval true request has been sent or was unnecessary
    \retval false sending the request failed because user has no address or the address is blocked */
bool
Transmitter::maybeRequestConfirmation(string_t user)
{
    string_t authUser;
    string_t smtpAddress;
    try {
        resolveAddress("user:" + user, smtpAddress, authUser);
        return true;
    }
    catch (...) {
        return false;
    }
}

bool
Transmitter::checkConfig(string_t key, string_t value)
{
    if (key == "WWW.KEY") {
        system_key = value;
        return true;
    } else if (key == "WWW.URL") {
        web_root = value;
        return true;
    } else if (key == "MAILOUT.MAXAGE") {
        /* @q Mailout.MaxAge:Int (Config)
           Maximum age of a message, in minutes.
           A message that could not been sent for this time is dropped. */
        stringValue(value, max_msg_age);
        return true;
    } else {
        return false;
    }
}

/** Thread entry point. */
void*
Transmitter::entry(void* self)
{
    Transmitter& me = *static_cast<Transmitter*>(self);
    while (1) {
        me.work_sem.down();
        me.processWork();
    }
}

/** Process one item from the work queue. */
void
Transmitter::processWork()
{
    // Fetch message id
    int32_t mid;
    {
        MutexGuard g(queue_mutex);
        if (work_queue.empty()) {
            return;
        }
        mid = work_queue.front();
    }

    // Process message
    DbSubtree root(database_connection, MAIL_ROOT);
    std::auto_ptr<Message> msg;
    StringList_t receivers;
    {
        // Obtain message object
        ReconnectingGuard g(database_rc);
        msg.reset(new Message(root, mid, "sending"));

        // Still active?
        bool active = true;
        string_t uid = msg->uniqueId().get();
        if (!uid.empty() && root.hashKey("uniqid").getIntField(uid) != mid) {
            lockedPrintf("[transmit] [msg%d] expired (replaced by new instance)\n", int(mid));
            active = false;
        }

        if (getCurrentTime() > msg->expireTime().get()) {
            lockedPrintf("[transmit] [msg%d] expired (too old)\n", int(mid));
            active = false;
        }

        if (!active) {
            msg->remove();
            {
                MutexGuard g(queue_mutex);
                work_queue.remove(mid);
            }
            return;
        }

        // Get receivers
        msg->receivers().getAll(receivers);
    }

    // Send it
    bool keep = false;
    for (uint32_t i = 0; i < receivers.size(); ++i) {
        bool keepThis = false;
        try {
            if (!sendMessage(*msg, receivers[i])) {
                keepThis = true;
                lockedPrintf("[transmit] [msg%d] receiver '%s' postponed\n", int(mid), receivers[i].c_str());
            } else {
                lockedPrintf("[transmit] [msg%d] receiver '%s' succeeded\n", int(mid), receivers[i].c_str());
            }
        }
        catch (FileProblemException& e) {
            lockedPrintf("[transmit] [msg%d] receiver '%s' failed: %s: %s\n", int(mid), receivers[i].c_str(), e.getFileName().c_str(), e.what());
        }
        catch (std::exception& e) {
            lockedPrintf("[transmit] [msg%d] receiver '%s' failed: %s\n", int(mid), receivers[i].c_str(), e.what());
        }
        if (!keepThis) {
            // Message failed or succeeded: remove from database
            ReconnectingGuard g(database_rc);
            msg->receivers().remove(receivers[i]);
        } else {
            // Message postponed: keep it
            keep = true;
        }
    }

    // Postprocess
    if (keep) {
        // Keep message because it has unverified addresses
        lockedPrintf("[transmit] [msg%d] keeping\n", int(mid));
        {
            MutexGuard g(queue_mutex);
            work_queue.remove(mid);
            pend_queue.push_back(mid);
        }
    } else {
        // Discard message because it has been sent or permanently failed
        lockedPrintf("[transmit] [msg%d] finished\n", int(mid));
        {
            ReconnectingGuard g(database_rc);
            msg->remove();
        }
        {
            MutexGuard g(queue_mutex);
            work_queue.remove(mid);
        }
    }
}

bool
Transmitter::sendMessage(Message& msg, string_t address)
{
    // Resolve email address
    string_t smtpAddress;
    string_t authUser;
    if (!resolveAddress(address, smtpAddress, authUser)) {
        return false;
    }

    // Prepare message
    string_t tplName;
    Template tpl;
    tpl.addVariable("SMTP_FROM", Smtp::getFrom());
    tpl.addVariable("SMTP_FQDN", Smtp::getFqdn());
    tpl.addVariable("SMTP_TO", smtpAddress);
    tpl.addVariable("USER", authUser);
    tpl.addVariable("CGI_ROOT", web_root);
    {
        ReconnectingGuard g(database_rc);
        {
            StringList_t args;
            msg.arguments().getAll(args);
            for (uint32_t i = 0; i+1 < args.size(); i += 2) {
                tpl.addVariable(args[i], args[i+1]);
            }
        }
        {
            StringList_t atts;
            msg.attachments().getAll(atts);
            for (uint32_t i = 0; i < atts.size(); ++i) {
                tpl.addFile(atts[i]);
            }
        }
        tplName = msg.templateName().get();
    }

    // Generate
    Ptr<Stream> s = Stream::openFile(Directory::makeFileName(template_dir, tplName), Stream::C_READ);
    TextFile tf(*s);
    std::auto_ptr<MimeBuilder> smtpMessage(tpl.generate(tf, authUser, smtpAddress));

    // Send
    Smtp().sendMail(Smtp::getFrom(), smtpAddress, *smtpMessage);

    return true;
}

/** Resolve an address.
    \param address [in] Address from message queue
    \param smtpAddress [out] Resolved SMTP address if any
    \retval true Address resolved correctly, send the message
    \retval false Address temporarily failed, postpone the message
    \throw std::runtime_error Address permanently failed, discard the message */
bool
Transmitter::resolveAddress(string_t address, string_t& smtpAddress, string_t& authUser)
{
    ReconnectingGuard g(database_rc);
    DbSubtree emailRoot(database_connection, EMAIL_ROOT);

    if (address.size() > 5 && address.compare(0, 5, "mail:", 5) == 0) {
        // directly specified SMTP address
        smtpAddress.assign(address, 5, string_t::npos);
        authUser = "anon";

        // check blocked address
        DbHash emailInfo(emailRoot.subtree(smtpAddress).hashKey("status"));
        if (emailInfo.getStringField("status/" + authUser) == "b") {
            throw std::runtime_error(string_t(format("email address '%s' is blocked", smtpAddress)).c_str());
        }
        return true;
    } else if (address.size() > 5 && address.compare(0, 5, "user:", 5) == 0) {
        // user Id
        string_t user(address, 5);

        // Fetch user email
        DbSubtree userRoot(database_connection, USER_ROOT);
        string_t userEmail = userRoot.subtree(user).hashKey("profile").getStringField("email");
        if (userEmail.empty()) {
            throw std::runtime_error(string_t(format("user '%s' has no email address", user)).c_str());
        }

        // Check email status
        DbHash emailInfo(emailRoot.subtree(userEmail).hashKey("status"));
        string_t emailStatus = emailInfo.getStringField("status/" + user);
        if (emailStatus.empty() || emailStatus == "u") {
            // Unconfirmed. Request confirmation.
            requestConfirmation(user, userEmail);
            return false;
        } else if (emailStatus == "r") {
            // Confirmation is requested. Check for expiration.
            int32_t expire = emailInfo.getIntField("expire/" + user);
            if (expire != 0 && getCurrentTime() > expire) {
                requestConfirmation(user, userEmail);
            }
            return false;
        } else if (emailStatus == "c") {
            // Address is confirmed, so return it
            smtpAddress = userEmail;
            authUser = user;
            return true;
        } else {
            // treat as blocked
            throw std::runtime_error(string_t(format("user '%s's email address '%s' is blocked", user, userEmail)).c_str());
        }
    } else {
        throw std::runtime_error(string_t(format("invalid address '%s'", address)).c_str());
    }
}

/** Request confirmation for an email address. Called with database lock held. */
void
Transmitter::requestConfirmation(string_t user, string_t userEmail)
{
    // Mark that we requested confirmation
    DbHash emailInfo(DbSubtree(database_connection, EMAIL_ROOT).subtree(userEmail).hashKey("status"));
    emailInfo.setField("status/" + user, "r");
    emailInfo.setField("expire/" + user, getCurrentTime() + 200*24*60 /* 200 days */);

    string_t userName = DbSubtree(database_connection, USER_ROOT).subtree(user).stringKey("name").get();

    // Create message
    std::auto_ptr<Message> msg(allocateMessage(DbSubtree(database_connection, MAIL_ROOT)));
    msg->uniqueId().set("confirmation-" + userEmail);
    msg->templateName().set("confirm");
    msg->arguments().setField("email", userEmail);
    msg->arguments().setField("user", userName);
    msg->arguments().setField("confirmlink", web_root + "confirm.cgi?key="
                              + encodeUrl(getConfirmLink(user, userEmail)) + "&mail="
                              + encodeUrl(userEmail));
    msg->receivers().add("mail:" + userEmail);

    // Log it
    lockedPrintf("[transmit] [msg%d] confirmation request for '%s', user '%s' queued\n", int(msg->getId()), userEmail.c_str(), user.c_str());
    msg->send();
    send(msg->getId());
}

/** Get current time. We store MINUTES since epoch, giving us a little longer
    than 2038 if time() works right. */
int32_t
getCurrentTime()
{
    return time(0) / 60;
}

/** Get maximum age of a message in minutes. */
int32_t
getMaximumMessageAge()
{
    return max_msg_age;
}

/** Confirm an email address. */
bool
confirmMail(string_t mail, string_t key, string_t info)
{
    // Find user Id
    string_t decoded = decodeBase64(key);
    string_t::size_type n = decoded.find(',');
    if (n == 0 || n >= decoded.size()) {
        lockedPrintf("[transmit] [auth] request for '%s' is syntactically invalid\n", mail.c_str());
        return false;
    }
    string_t user(decoded, 0, n);

    // Check hash
    if (key != getConfirmLink(user, mail)) {
        lockedPrintf("[transmit] [auth] request for '%s' lacks proper signature\n", mail.c_str());
        return false;
    }

    // OK, operate
    ReconnectingGuard g(database_rc);
    DbHash emailInfo(DbSubtree(database_connection, EMAIL_ROOT).subtree(mail).hashKey("status"));
    emailInfo.setField("status/" + user, "c");
    if (info.empty()) {
        emailInfo.removeField("confirm/" + user);
    } else {
        emailInfo.setField("confirm/" + user, info);
    }
    lockedPrintf("[transmit] [auth] request for '%s' user '%s' accepted\n", mail.c_str(), user.c_str());
    return true;
}
